/**
 * @file      arm9-isa.cpp
 * @author    Marcio Juliato
 *            Alexandro Baldassin
 *            
 * update     Kleber Silva Rodrigues
 *
 *            The ArchC Team
 *            http://www.archc.org/
 *
 *            Computer Systems Laboratory (LSC)
 *            IC-UNICAMP
 *            http://www.lsc.ic.unicamp.br
 *
 * @version   1.0
 * @date      Mon, 19 Jun 2006 15:33:28 -0300
 * 
 * @brief     The ArchC R3000 cycle-accurate model.
 * 
 * @attention Copyright (C) 2002-2006 --- The ArchC Team
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "archc.H"
//#include "math.h"
#include "ac_types.H"
#include "ac_parms.H"
#include  "arm9-isa.H"
#include  "ac_isa_init.cpp"

//ac_Uword CPSR;
ac_Uword data1, data2, quantidadeShift, mascara;


/************************************************************************/
//!Generic instruction behavior method.
void ac_behavior( instruction ){

  switch( stage ) {
  case IF:
    ac_pc += 4;
    IF_ID.instruction = ac_pc;
    break;
    
  case ID:

	/** verifica o campo 'cond' da instrucao **/
	switch(cond){
	
		case 0x00: //EQ - Equal - Z set
			if(CPSR.z == 0x01)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x01: //NE - Not Equal - Z clear
			if(CPSR.z == 0x00)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x02: //CS/HS - Carry Set/Unsigned higher or same - C set
			if(CPSR.c == 0x01)	fID_EX.lag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x03: //CC/LO - Carry Clear/Unsigned lower - C clear
			if(CPSR.c == 0x00)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x04: //MI - Minus/Negative - N set
			if(CPSR.n == 0x01)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x05: //PL- Plus/Positive or zero - N clear
			if(CPSR.n == 0x00)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x06: //VL - Overflow - V set
			if(CPSR.v == 0x01)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x07: //VC - No Overflow - V clear
			if(CPSR.v == 0x00)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x08: //HI - Unsigned Higher - C set and Z clear
			if(CPSR.c == 0x01 && CPSR.z = 0x00)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x09: //LS - Unsigned Lower - C clear or Z set
			if(CPSR.c == 0x00 || CPSR.z == 0x01)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x0a: //GE - Signed Greather than or Equal - N set and V set or N clear and V clear (N == V)
			if(CPSR.n == CPSR.v)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x0b: //LT - Signed Less than - N set and V clear or N clear and V set (N != V)
			if(CPSR.n != CPSR.v)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x0c: //GT - Signed Greather than - Z clear, and either N set and V set, or N clear and v clear (Z == 0 && N == V) 
			if(CPSR.z == 0 && CPSR.n == CPSR.v)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x0d: //LE - Signed Less than or equal -  Z clear, or N set and V clear, or N clear and v set (Z == 1 || N != V) 
			if(CPSR.z == 1 || CPSR.n != CPSR.v)	ID_EX.flag_cond = 0x01;	else	ID_EX.flag_cond = 0x00;
		break;
		case 0x0e: //AL - Always (Unconditional)
			ID_EX.flag_cond = 0x01;
		break;
		case 0x0f: //NV - Never
			// Depende da versao. Geralmente eh UNPERDICTABLE
		break;
	}	

    break;

  case EX:

    break;

  case MEM:
    break;

  case WB:
    /* Execute write back when allowed */
    if (MEM_WB.regwrite == 1) {
      // Register 0 is never written
      if (MEM_WB.rdest != 0)
	RB.write(MEM_WB.rdest, MEM_WB.wbdata);
    }
    break;

  default:
    break;
  }
};
/************************************************************************/


/************************************************************************/
//! Common behaviour for all R Type instructions
void ac_behavior( Type_DT ){

  switch( stage ) {
  case IF:
    break;

  case ID:
    /* Assuming that all DT type instructions execute write back. When it is not necessary,
       the instruction BY ITSELF must clear the regwrite bit on its ID stage */
    ID_EX.regwrite = 1;
    ID_EX.memread = 0;
    ID_EX.memwrite = 0;
    ID_EX.instruction = IF_ID.instruction;	// le instrucao
    ID_EX.data1 = RB.read(rn);	// le o conteudo do primeiro registrador
    ID_EX.rd = rd;	// le o numero de registrador de destino
   
   break; 
  	case EX:

	// faz os deslocamentos e verifica quem eh o segundo operando
	if( immediate_flag == 0x01){
		// ele sempre vai rotacionar o imediato p/ esquerda , no caso tem que deixar o campo rotate = 0:
		data2 = ((immediate<<rotate)|(immediate>>(8-rotate)));	
	}
	else{
	// se shift_reg_flag == 1 , a quantidade de shifts eh dada por registrador; se == 0, eh dada por imediato:
	if( shift_reg_flag == 0x01)	quantidadeShift = RB.read(rs);
        	else 	quantidadeShift = shiffted_imm;
	// aplicar o shift
	if( shif_type == 0x00) data2 = RB.read(rm) << quantidadeShift;	// desloca para esq
		else if( shif_type == 0x01)	data2 = RB.read(rm) >> quantidadeShift;	// desloca para dir
		else if( shif_type == 0x02){		// arithmetic right
			
			if(RB.read(rm) < 0x00){
				mascara = 0xffff;
				data2 = (RB.read(rm) >> quantidadeShift) | ( mascara << (8 - quantidadeShift));
				} else 	data2 = RB.read(rm) >> quantidadeShift;
				
			}
			else if( shif_type == 0x03){
				if( shift_reg_flag == 0x00 && quantidadeShift == 0x00 ){	// aplicar RRX (rotate extendido)
					for(int i = 0; i < quantidadeShift; i++){
						mascara = (0x00 + CPSR.c) << 31;	// obtem o bit de CPSR.c
						data2 = ( RB.read(rm) >> 1 ) + mascara;	// define o valor do registrador
						CPSR.c =  RB.read(rm) & 0x01; 		// guarda o bit de rotate no CPSR.c
					} 
				}
				else	data2 = ((RB.read(rm)>>quantidadeShift)|(RB.read(rm)<<(32-quantidadeShift)));	// aplicar ROR (rotate normal)
			}
		}
	
	break;

  case MEM:
    /* Just copies the values from EX_MEM to MEM_WB */
    MEM_WB.regwrite = EX_MEM.regwrite;
    MEM_WB.wbdata = EX_MEM.alures;
    MEM_WB.rdest = EX_MEM.rdest;
    break;

  case WB:
    break;

  default:
    break;
  }
}
/************************************************************************/

//!Instruction add behavior method.
void ac_behavior( MOV ){
  
  switch( stage ) {
  case IF:
    /* It was implemented for all instructions on ac_behaviour( instruction ) */
    break;

  case ID:
    break;

  case EX:

	// A instrucao soh pode ser executada se passar pelo teste de cond
	if (ID_EX.flag_cond == 0x01){
		EX_MEM.alures = ID_EX.data1+ data2;
		printf("MOV: %d\n",ID_EX.data1+ data2 );
	}

 	EX_MEM.rdest = ID_EX.rd;		// valor do registrador de destino
    	EX_MEM.regwrite = ID_EX.regwrite;	// sinal controle escrita reg
    	EX_MEM.memread = ID_EX.memread;		// sinal comrtole leitura memoria
	EX_MEM.memwrite = ID_EX.memwrite;	// sinal controle escrita memoria
   
    break;

  case MEM:
    break;

  case WB:
    break;

  default:
    break;
  }
};


//!Instruction add behavior method.
void ac_behavior( ADD ){
  
  switch( stage ) {
  case IF:
    /* It was implemented for all instructions on ac_behaviour( instruction ) */
    break;

  case ID:
    break;

  case EX:

	// A instrucao executa apos passar pelo teste de cond
	if (ID_EX.flag_cond == 0x01){
    		EX_MEM.alures = ID_EX.data1+ data2;
		printf("ADD: %d\n",ID_EX.data1+ data2 ); // mostra o resultado
		if(set_flags == 0x01){
			CPSR.n = EX_MEM.alures >> 31;
			if(EX_MEM.alures = 0)
			CPSR.z = 0x01;
			else CPSR.z = 0x00;
			// verifica se tem valor no data1/data2
			if((ID_EX.data1 >= 0x00 && data2 >= 0x00)) {
		

				CPSR.c = 1;
			}else 	CPSR.c = 0;
			  /* Teste Overflow */
			  if ( ((ID_EX.data1 & 0x80000000) == (EX_MEM.alures & 0x80000000)) &&
			       ((EX_MEM.alures & 0x80000000) != (data2 & 0x80000000)) ) {
			       fprintf(stderr, "EXCEPTION(add): integer overflow.\n"); exit(EXIT_FAILURE);
			    }
		}
	}
	

	
 	EX_MEM.rdest = ID_EX.rd;		// valor do registrador de destino
    	EX_MEM.regwrite = ID_EX.regwrite;	// sinal controle escrita reg
    	EX_MEM.memread = ID_EX.memread;		// sinal comrtole leitura memoria
	EX_MEM.memwrite = ID_EX.memwrite;	// sinal controle escrita memoria
   
    break;

  case MEM:
    break;

  case WB:
    break;

  default:
    break;
  }
};


//!Instruction and behavior method.
void ac_behavior( AND ){

// nao deu tempo :(

}
